/*
 * Simple FUSE server testing harness.
 *
 
        Copyright (c) 2006 Russ Cox.

Permission to use, copy, modify, and distribute this software for any
purpose without fee is hereby granted, provided that this entire notice
is included in all copies of any software which is or includes a copy
or modification of this software and in all copies of the supporting
documentation for such software.

THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR IMPLIED
WARRANTY.  IN PARTICULAR, THE AUTHOR MAKES NO REPRESENTATION OR WARRANTY
OF ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
FITNESS FOR ANY PARTICULAR PURPOSE.
*/
 
#define _GNU_SOURCE 1	/* for O_DIRECTORY on Linux */
#include "a.h"
#include <fcntl.h>
#include <time.h>
#include "arg.h"

/* GNUisms */
#ifndef O_DIRECTORY
#define O_DIRECTORY 0
#endif
#ifndef O_LARGEFILE
#if defined(__linux__)
#define O_LARGEFILE 0100000  /* Sigh */
#else
#define O_LARGEFILE 0
#endif
#endif

void fusehandle(FuseMsg*);

int debug;
char *argv0;
char *mtpt;
char *mode;

/*
 * The number of seconds that the kernel can cache
 * returned file attributes.  FUSE's default is 1.0.
 * I haven't experimented with using 0.
 */
double attrtimeout = 1.0;

/*
 * The number of seconds that the kernel can cache
 * the returned entry nodeids returned by lookup.
 * I haven't experimented with other values.
 */
double entrytimeout = 1.0;

void
usage(void)
{
	fprint(2, "usage: fusetest [-D] [-a attrtimeout] [-e entrytimeout] mode mtpt\n");
	fprint(2, "mode is one of:\n");
	fprint(2, "  empty - empty file system\n");
	fprint(2, "  simple - simple 1-file file system\n");
	fprint(2, "  huge - enormous 2^50-file file system\n");
	fprint(2, "  timeout - huge, but stop answering requests after 2 seconds\n");
	fprint(2, "  many - mount as many fuses as possible,\n");
	fprint(2, "    creating subdirectories of mtpt as needed\n");
	exit(1);
}

void empty(void);
void simple(void);
void timeout(void);
void huge(void);
void manymount(void);

int
main(int argc, char **argv)
{
	ARGBEGIN{
	case 'D':
		debug++;
		break;
	case 'a':
		attrtimeout = atof(EARGF(usage()));
		break;
	case 'e':
		entrytimeout = atof(EARGF(usage()));
		break;
	default:
		usage();
	}ARGEND

	if(argc != 2)
		usage();

	quotefmtinstall();
	fmtinstall('G', fusefmt);
	mode = argv[0];
	mtpt = argv[1];

	if(strcmp(mode, "many") == 0){
		manymount();
		return 0;
	}

	initfuse(argv[1]);
	if(fork() == 0){
		FuseMsg *m;

		while((m = readfusemsg()) != nil)
			fusehandle(m);

		fusemtpt = nil;	/* no need to unmount after EOF */
		return 0;
	}

	/*
	 * Now that we're serving FUSE, we can wait
	 * for the mount to finish and exit back to the user.
	 */
	waitfuse();
	return 0;
}

void
empty(void)
{
}

void
simple(void)
{
}

void
timeout(void)
{
}

void
huge(void)
{
}

int mountcount;

void unmountfuse(char*);

void
manycleanup(void)
{
	char buf[1024];

	print("mounted %d before failure; cleaning up...\n", mountcount);
	for(; mountcount>=0; mountcount--){
		snprint(buf, sizeof buf, "%s/mtpt%d", mtpt, mountcount);
		unmountfuse(buf);
		if(rmdir(buf) < 0)
			fprint(2, "rmdir %s: %r\n", buf);
	}	
}

void
manymount(void)
{
	char buf[1024];

	atexit(manycleanup);
	for(mountcount=0;; mountcount++){
		snprint(buf, sizeof buf, "%s/mtpt%d", mtpt, mountcount);
		if(access(buf, 0) < 0 && mkdir(buf, 0777) < 0)
			sysfatal("mkdir %s: %r", buf);
		initfuse(buf);	// will sysfatal on failure
	}
}

/*
 * FUSE uses nodeids to refer to active "struct inodes"
 * (9P's unopened fids).  FUSE uses fhs to refer to active
 * "struct fuse_files" (9P's opened fids).  The choice of 
 * numbers is up to us except that nodeid 1 is the root directory.
 * We use the same number space for both and call the 
 * bookkeeping structure a FuseFid.
 *
 * FUSE requires nodeids to have associated generation 
 * numbers.  If we reuse a nodeid, we have to bump the 
 * generation number to guarantee that the nodeid,gen
 * combination is never reused.
 * 
 * There are also inode numbers returned in directory reads
 * and file attributes, but these do NOT need to match the nodeids.
 */
typedef struct Fusefid Fusefid;
struct Fusefid
{
	Fusefid *next;
	char *name;
	int ref;
	int id;
	int gen;
	int isnodeid;
	
	/* directory read state */
	vlong off;
};

Fusefid **fusefid;
int nfusefid;
Fusefid *freefusefidlist;

Fusefid*
allocfusefid(void)
{
	Fusefid *f;
	
	if((f = freefusefidlist) == nil){
		f = emalloc(sizeof *f);
		fusefid = erealloc(fusefid, (nfusefid+1)*sizeof *fusefid);
		f->id = nfusefid;
		fusefid[f->id] = f;
		nfusefid++;
	}else
		freefusefidlist = f->next;
	f->next = nil;
	f->ref = 1;
	f->isnodeid = -1;
	return f;
}

void
freefusefid(Fusefid *f)
{
	if(--f->ref > 0)
		return;
	free(f->name);
	f->name = nil;
	f->off = 0;
	f->next = freefusefidlist;
	f->isnodeid = -1;
	freefusefidlist = f;
}

uvlong
_alloc(char *name, int isnodeid)
{
	Fusefid *ff;
	
	ff = allocfusefid();
	ff->isnodeid = isnodeid;
	ff->name = name;
	ff->gen++;
	return ff->id+2; /* skip 0 and 1 */
}

uvlong
allocfh(char *name)
{
	return _alloc(name, 0);
}
uvlong
allocnodeid(char *name)
{
	return _alloc(name, 1);
}

Fusefid*
lookupfusefid(uvlong id, int isnodeid)
{
	Fusefid *ff;
	if(id < 2 || id >= nfusefid+2)
		return nil;
	ff = fusefid[(int)id-2];
	if(ff->isnodeid != isnodeid)
		return nil;
	return ff;
}

char*
_lookupname(uvlong id, int isnodeid)
{
	Fusefid *ff;
	
	if((ff = lookupfusefid(id, isnodeid)) == nil)
		return nil;
	return ff->name;
}

char*
fh2name(uvlong fh)
{
	return _lookupname(fh, 0);
}

char*
nodeid2name(uvlong nodeid)
{
	if(nodeid == 1)
		return "/";
	return _lookupname(nodeid, 1);
}

void
name2attr(char *name, int isdir, struct fuse_attr *attr)
{
	attr->ino = (uint)name;
	attr->size = 0;
	attr->blocks = (attr->size+8191)/8192;
	attr->atime = time(0);
	attr->mtime = time(0);
	attr->ctime = time(0);
	attr->atimensec = 0;
	attr->mtimensec = 0;
	attr->ctimensec = 0;
	attr->mode = 0666;
	if(isdir)
		attr->mode |= 0111 | S_IFDIR;
	else
		attr->mode |= S_IFREG;
	attr->nlink = 1;
	attr->uid = getuid();
	attr->gid = getgid();
	attr->rdev = 0;
}

void
f2timeout(double f, __u64 *s, __u32 *ns)
{
	*s = f;
	*ns = (f - (int)f)*1e9;
}

void
name2attrout(char *name, int isdir, struct fuse_attr_out *out)
{
	f2timeout(attrtimeout, &out->attr_valid, &out->attr_valid_nsec);
	name2attr(name, isdir, &out->attr);
}

/*
 * Lookup.  Walk to the name given as the argument.
 * The response is a fuse_entry_out giving full stat info.
 */
void
fuselookup(FuseMsg *m)
{
	char *path, *name;
	Fusefid *ff;
	vlong x, y;
	struct fuse_entry_out out;
	
	path = m->tx;
	if((name = nodeid2name(m->hdr->nodeid)) == nil){
		replyfuseerrno(m, ESTALE);
		return;
	}
	if(path[0] == 0)
		goto bad;
	/*
	 * Figure out whether path is in the directory 'name'.
	 * If so goto have.
	 * See also fusereaddir, which has same logic.
	 */
	if(strcmp(mode, "simple") == 0){
		if(strcmp(path, "f") == 0)
			goto have;
	}
	if(strcmp(mode, "huge") == 0){
		x = atoll(name+1);
		y = atoll(path+1);
		if(path[0] == 'd' && x < (1LL<<50) && (y == 2*x+1 || y == 2*x+2))
			goto have;
		if(path[0] == 'f' && x >= (1LL<<50) && (y == 2*x+1 || y == 2*x+2))
			goto have;
	}
	if(strcmp(mode, "linear") == 0){
		x = atoll(name+1);
		y = atoll(path+1);
		if(path[0] == 'd' && x < (1LL<<50) && y == x+1)
			goto have;
		if(path[0] == 'f' && x >= (1LL<<50) && y == x+1)
			goto have;
	}
	if(strcmp(mode, "flat") == 0){
		y = atoll(path+1);
		if(path[0] == 'f' && y < (1LL<<50))
			goto have;
	}
bad:
	replyfuseerrno(m, ENOENT);
	return;

have:
	path = estrdup(path);
	out.nodeid = allocnodeid(path);
	ff = lookupfusefid(out.nodeid, 1);
	out.generation = ff->gen;
	f2timeout(attrtimeout, &out.attr_valid, &out.attr_valid_nsec);
	f2timeout(entrytimeout, &out.entry_valid, &out.entry_valid_nsec);
	name2attr(path, path[0] == 'd' || path[0] == '/', &out.attr);
	replyfuse(m, &out, sizeof out);
}

/*
 * Forget.  Reference-counted clunk for nodeids.
 * Does not send a reply.
 * Each lookup response gives the kernel an additional reference 
 * to the returned nodeid.  Forget says "drop this many references
 * to this nodeid".  Our fuselookup, when presented with the same query,
 * does not return the same results (it allocates a new nodeid for each
 * call), but if that ever changes, fuseforget already handles the ref
 * counts properly.
 */
void
fuseforget(FuseMsg *m)
{
	struct fuse_forget_in *in;
	Fusefid *ff;

	in = m->tx;
	if((ff = lookupfusefid(m->hdr->nodeid, 1)) == nil)
		return;
	if(ff->ref > in->nlookup){
		ff->ref -= in->nlookup;
		return;
	}
	if(ff->ref < in->nlookup)
		fprint(2, "bad count in forget\n");
	ff->ref = 1;
	freefusefid(ff);
}

/*
 * Getattr.
 * Replies with a fuse_attr_out structure giving the
 * attr for the requested nodeid in out.attr.
 * Out.attr_valid and out.attr_valid_nsec give 
 * the amount of time that the attributes can
 * be cached.
 *
 * Empirically, though, if I run ls -ld on the root
 * twice back to back, I still get two getattrs,
 * even with a one second attribute timeout!
 */
void
fusegetattr(FuseMsg *m)
{
	char *name;
	struct fuse_attr_out out;

	if((name = nodeid2name(m->hdr->nodeid)) == nil){
		replyfuseerrno(m, ESTALE);
		return;
	}
	memset(&out, 0, sizeof out);
	name2attrout(name, name[0]=='d' || name[0]=='/', &out);
	replyfuse(m, &out, sizeof out);
}

/*
 * Setattr.
 * FUSE treats the many Unix attribute setting routines
 * more or less like 9P does, with a single message.
 */
void
fusesetattr(FuseMsg *m)
{
	replyfuseerrno(m, EPERM);
}

char*
_fuseopenfid(uvlong nodeid, int isdir, int openmode, int *err)
{
	char *name;

	if((name = nodeid2name(nodeid)) == nil){
		*err = ESTALE;
		return nil;
	}
	if(isdir && name[0] != 'd' && name[0] != '/'){
		*err = ENOTDIR;
		return nil;
	}
	if(openmode != O_RDONLY && isdir){
		*err = EISDIR;
		return nil;
	}
	return estrdup(name);
}

/*
 * Open & Opendir.
 * Argument is a struct fuse_open_in.
 * The mode field is ignored (presumably permission bits)
 * and flags is the open mode.
 * Replies with a struct fuse_open_out.
 */
void
_fuseopen(FuseMsg *m, int isdir)
{
	struct fuse_open_in *in;
	struct fuse_open_out out;
	int openmode, flags, err;
	char *nname;

	in = m->tx;
	flags = in->flags;
	openmode = flags&3;
	flags &= ~3;
	flags &= ~(O_DIRECTORY|O_NONBLOCK|O_LARGEFILE);
	if(flags & O_TRUNC){
		openmode |= O_TRUNC;
		flags &= ~O_TRUNC;
	}
	/*
	 * Could translate but not standard 9P:
	 *	O_DIRECT -> ODIRECT
	 *	O_NONBLOCK -> ONONBLOCK
	 *	O_APPEND -> OAPPEND
	 */
	if(flags){
		fprint(2, "unexpected open flags %#uo", (uint)in->flags);
		replyfuseerrno(m, EACCES);
		return;
	}
	if((nname = _fuseopenfid(m->hdr->nodeid, isdir, openmode, &err)) == nil){
		replyfuseerrno(m, err);
		return;
	}
	out.fh = allocfh(nname);
	out.open_flags = FOPEN_DIRECT_IO;	/* no page cache */	
	replyfuse(m, &out, sizeof out);
}

void
fuseopen(FuseMsg *m)
{
	_fuseopen(m, 0);
}

void
fuseopendir(FuseMsg *m)
{
	_fuseopen(m, 1);
}

/*
 * Create & Mkdir.
 */
void
fusemkdir(FuseMsg *m)
{
	replyfuseerrno(m, EPERM);
}

void
fusecreate(FuseMsg *m)
{
	replyfuseerrno(m, EPERM);
}

/*
 * Access.  (All is permitted.)
 */
void
fuseaccess(FuseMsg *m)
{
	replyfuse(m, nil, 0);
}

/*
 * Release.
 * Equivalent of clunk for file handles.
 * in->flags is the open mode used in Open or Opendir.
 */
void
fuserelease(FuseMsg *m)
{
	struct fuse_release_in *in;
	Fusefid *ff;
	
	in = m->tx;
	if((ff = lookupfusefid(in->fh, 0)) != nil)
		freefusefid(ff);
	else
		fprint(2, "fuserelease: fh not found\n");
	replyfuse(m, nil, 0);
}

void
fusereleasedir(FuseMsg *m)
{
	fuserelease(m);
}

/*
 * Read.
 * Read from file handle in->fh at offset in->offset for size in->size.
 * We truncate size to maxwrite just to keep the buffer reasonable.
 */
void
fuseread(FuseMsg *m)
{
	int n;
	uchar *buf;
	char *name;
	struct fuse_read_in *in;

	in = m->tx;
	if((name = fh2name(in->fh)) == nil){
		replyfuseerrno(m, ESTALE);
		return;
	}
	n = in->size;
	if(n > fusemaxwrite)
		n = fusemaxwrite;
	buf = emalloc(n);

	/* Pretend each file contains its name. */
	if(n > strlen(name))
		n = strlen(name);
	if(in->offset >= n || in->offset < 0)
		n = 0;
	else{
		n -= in->offset;
		memmove(buf, name+in->offset, n);
	}

	replyfuse(m, buf, n);
	free(buf);
}

/* 
 * Readdir.
 * Read from file handle in->fh at offset in->offset for size in->size.
 * We truncate size to maxwrite just to keep the buffer reasonable.
 * If it became necessary, we could implement a crude seek
 * or cache the entire list of directory entries.
 */
int canpack(char*, uvlong, uchar**, uchar*);
void
fusereaddir(FuseMsg *m)
{
	struct fuse_read_in *in;
	uchar *buf, *p, *ep;
	int n;
	vlong x;
	char *name, tmp[40];
	Fusefid *ff;
	
	in = m->tx;
	if((ff = lookupfusefid(in->fh, 0)) == nil){
		replyfuseerrno(m, ESTALE);
		return;
	}
	n = in->size;
	if(n > fusemaxwrite)
		n = fusemaxwrite;
	buf = emalloc(n);
	p = buf;
	ep = buf + n;
	for(;;){
		// Figure out name at ff->off or break.
		if(strcmp(mode, "simple") == 0 && ff->off == 0)
			name = "f";
		else if(strcmp(mode, "huge") == 0 && ff->off < 2){
			/*
			 * Huge tree in heap order:
			 * all directories dx (x in decimal)
			 * contain d(2x+1) and d(2x+2).
			 * The bottom level contains files instead of directories.
			 */
			x = atoll(ff->name+1);
			snprint(tmp, sizeof tmp, "d%lld", 2*x+1+ff->off);
			if(x >= (1LL<<50))
				tmp[0] = 'f';
			name = tmp;
		}
		else if(strcmp(mode, "linear") == 0 && ff->off < 1){
			/* 
			 * Huge linear tree.
			 * Directory dx contains d(x+1).
			 */
			 x = atoll(ff->name+1);
			 snprint(tmp, sizeof tmp, "d%lld", x+1);
			 if(x >= (1LL<<50))
			 	tmp[0] = 'f';
			 name = tmp;
		}
		else if(strcmp(mode, "flat") == 0 && ff->off < (1LL<<50)){
			/*
			 * Huge flat tree.
			 * Directory / contains 2^50 files.
			 */
			snprint(tmp, sizeof tmp, "f%lld", ff->off);
			name = tmp;
		}
		else
			break;
		if(!canpack(name, ff->off, &p, ep))
			goto out;
		ff->off++;
	}
out:			
	replyfuse(m, buf, p - buf);
	free(buf);
}

int
canpack(char *name, uvlong off, uchar **pp, uchar *ep)
{
	uchar *p;
	struct fuse_dirent *de;
	int pad, size;
	
	p = *pp;
	size = FUSE_NAME_OFFSET + strlen(name);
	pad = 0;
	if(size%8)
		pad = 8 - size%8;
	if(size+pad > ep - p)
		return 0;
	de = (struct fuse_dirent*)p;
	de->ino = (uint)name;
	de->off = off;
	de->namelen = strlen(name);
	memmove(de->name, name, de->namelen);
	if(pad > 0)
		memset(de->name+de->namelen, 0, pad);
	*pp = p+size+pad;
	return 1;
}

/*
 * Write.
 * Write from file handle in->fh at offset in->offset for size in->size.
 * Don't know what in->write_flags means.
 * 
 * Apparently implementations are allowed to buffer these writes
 * and wait until Flush is sent, but FUSE docs say flush may be
 * called zero, one, or even more times per close.  So better do the
 * actual writing here.  Also, errors that happen during Flush just
 * show up in the close() return status, which no one checks anyway.
 */
void
fusewrite(FuseMsg *m)
{
	replyfuseerrno(m, EPERM);
}

/*
 * Flush.  Supposed to flush any buffered writes.  Don't use this.
 * 
 * Flush is a total crock.  It gets called on close() of a file descriptor
 * associated with this open file.  Some open files have multiple file
 * descriptors and thus multiple closes of those file descriptors.
 * In those cases, Flush is called multiple times.  Some open files
 * have file descriptors that are closed on process exit instead of
 * closed explicitly.  For those files, Flush is never called.
 * Even more amusing, Flush gets called before close() of read-only
 * file descriptors too!
 * 
 * This is just a bad idea.
 */
void
fuseflush(FuseMsg *m)
{
	replyfuse(m, nil, 0);
}

/*
 * Unlink & Rmdir.
 */
void
fuseunlink(FuseMsg *m)
{
	replyfuseerrno(m, EPERM);
}

void
fusermdir(FuseMsg *m)
{
	replyfuseerrno(m, EPERM);
}

void
fuserename(FuseMsg *m)
{
	replyfuseerrno(m, EPERM);
}

/*
 * Fsync.  Commit file info to stable storage.
 * Not sure what in->fsync_flags are.
 */
void
fusefsync(FuseMsg *m)
{
	replyfuse(m, nil, 0);
}

/*
 * Fsyncdir.  Commit dir info to stable storage?
 */
void
fusefsyncdir(FuseMsg *m)
{
	fusefsync(m);
}

/*
 * Statfs.  Send back information about file system.
 * Not really worth implementing, except that if we
 * reply with ENOSYS, programs like df print messages like
 *   df: `/tmp/z': Function not implemented
 * and that gets annoying.  Returning all zeros excludes
 * us from df without appearing to cause any problems.
 */
void
fusestatfs(FuseMsg *m)
{
	struct fuse_statfs_out out;
	
	memset(&out, 0, sizeof out);
	replyfuse(m, &out, sizeof out);
}

void (*fusehandlers[100])(FuseMsg*);

struct {
	int op;
	void (*fn)(FuseMsg*);
} fuselist[] = {
	{ FUSE_LOOKUP,		fuselookup },
	{ FUSE_FORGET,		fuseforget },
	{ FUSE_GETATTR,		fusegetattr },
	{ FUSE_SETATTR,		fusesetattr },
	/*
	 * FUSE_READLINK, FUSE_SYMLINK, FUSE_MKNOD are unimplemented.
	 */
	{ FUSE_MKDIR,		fusemkdir },
	{ FUSE_UNLINK,		fuseunlink },
	{ FUSE_RMDIR,		fusermdir },
	{ FUSE_RENAME,		fuserename },
	/*
	 * FUSE_LINK is unimplemented.
	 */
	{ FUSE_OPEN,		fuseopen },
	{ FUSE_READ,		fuseread },
	{ FUSE_WRITE,		fusewrite },
	{ FUSE_STATFS,		fusestatfs },
	{ FUSE_RELEASE,		fuserelease },
	{ FUSE_FSYNC,		fusefsync },
	/*
	 * FUSE_SETXATTR, FUSE_GETXATTR, FUSE_LISTXATTR, and
	 * FUSE_REMOVEXATTR are unimplemented. 
	 * FUSE will stop sending these requests after getting
	 * an -ENOSYS reply (see dispatch below).
	 */
	{ FUSE_FLUSH,		fuseflush },
	/*
	 * FUSE_INIT is handled in initfuse and should not be seen again.
	 */
	{ FUSE_OPENDIR,		fuseopendir },
	{ FUSE_READDIR,		fusereaddir },
	{ FUSE_RELEASEDIR,	fusereleasedir },
	{ FUSE_FSYNCDIR,	fusefsyncdir },
	{ FUSE_ACCESS,		fuseaccess },
	{ FUSE_CREATE,		fusecreate },
};

void
fusehandle(FuseMsg *m)
{
	int i;
	static int init;
	if(!init){
		init = 1;
		for(i=0; i<nelem(fuselist); i++){
			if(fuselist[i].op >= nelem(fusehandlers))
				sysfatal("make fusehandlers bigger op=%d", fuselist[i].op);
			fusehandlers[fuselist[i].op] = fuselist[i].fn;
		}
	}

	if((uint)m->hdr->opcode >= nelem(fusehandlers) 
	|| !fusehandlers[m->hdr->opcode]){
		replyfuseerrno(m, ENOSYS);
		return;
	}
	fusehandlers[m->hdr->opcode](m);
}
